from gluonts.dataset.common import ListDataset
from gluonts.dataset.field_names import FieldName
from gluonts.model.deepar import DeepAREstimator
from gluonts.mx.distribution.neg_binomial import NegativeBinomialOutput
from gluonts.mx.trainer import Trainer
from gluonts.evaluation.backtest import make_evaluation_predictions
from tqdm import tqdm
from htsmodels.results.calculate_metrics import calculate_metrics
import numpy as np
import pickle
from pathlib import Path


class DeepAR:

    def __init__(self, dataset, groups, input_dir='./'):
        self.dataset = dataset
        self.groups = groups
        self.input_dir = input_dir
        self._create_directories()
        self.stat_cat_cardinalities = [v for k, v in self.groups['train']['groups_n'].items()]
        self.stat_cat = np.concatenate(([v.reshape(-1, 1) for k, v in self.groups['train']['groups_idx'].items()]), axis=1)
        self.dates = [groups['dates'][0] for _ in range(groups['train']['s'])]

        time_interval = (self.groups['dates'][1] - self.groups['dates'][0]).days
        if time_interval < 8:
            self.time_int = 'W'
        elif time_interval < 32:
            self.time_int = 'M'
        elif time_interval < 93:
            self.time_int = 'Q'
        elif time_interval < 367:
            self.time_int = 'Y'

    def _create_directories(self):
        # Create directory to store results if does not exist
        Path(f'{self.input_dir}results').mkdir(parents=True, exist_ok=True)

    def _build_train_ds(self):
        train_target_values = self.groups['train']['data'].T

        train_ds = ListDataset([
            {
                FieldName.TARGET: target,
                FieldName.START: start,
                FieldName.FEAT_STATIC_CAT: fsc
            }
            for (target, start, fsc) in zip(train_target_values,
                                            self.dates,
                                            self.stat_cat)
        ], freq=self.time_int)

        return train_ds

    def _build_test_ds(self):
        test_target_values = self.groups['predict']['data'].reshape(self.groups['predict']['s'], self.groups['predict']['n'])

        test_ds = ListDataset([
            {
                FieldName.TARGET: target,
                FieldName.START: start,
                FieldName.FEAT_STATIC_CAT: fsc
            }
            for (target, start, fsc) in zip(test_target_values,
                                            self.dates,
                                            self.stat_cat)
        ], freq=self.time_int)

        return test_ds

    def train(self, lr=1e-3, epochs=100):
        train_ds = self._build_train_ds()

        estimator = DeepAREstimator(
            prediction_length=self.groups['h'],
            freq=self.time_int,
            distr_output=NegativeBinomialOutput(),
            use_feat_dynamic_real=False,
            use_feat_static_cat=True,
            cardinality=self.stat_cat_cardinalities,
            trainer=Trainer(
                learning_rate=lr,
                epochs=epochs,
                num_batches_per_epoch=50,
                batch_size=32
            )
        )

        model = estimator.train(train_ds)
        return model

    def predict(self, model):
        test_ds = self._build_test_ds()

        forecast_it, ts_it = make_evaluation_predictions(
            dataset=test_ds,
            predictor=model,
            num_samples=100
        )

        print("Obtaining time series predictions ...")
        forecasts = list(tqdm(forecast_it, total=len(test_ds)))

        return forecasts

    def results(self, forecasts, n_samples=100):
        res = np.zeros((len(forecasts), n_samples, self.groups['h']))
        for i, j in enumerate(forecasts):
            res[i] = j.samples

        res = np.concatenate((np.zeros((self.groups['train']['s'], n_samples, self.groups['train']['n']), dtype=np.float64), res), axis=2)
        res = np.transpose(res, (1, 2, 0))
        return res

    def store_metrics(self, res):
        with open(f'{self.input_dir}results/results_gp_cov_{self.dataset}.pickle', 'wb') as handle:
            pickle.dump(res, handle, pickle.HIGHEST_PROTOCOL)

    def metrics(self, mean):
        res = calculate_metrics(mean, self.groups)
        return res


__version__ = "0.1.4"

from htsmodels import models
from htsmodels import results

# Only print in interactive mode
import __main__ as main
if not hasattr(main, '__file__'):
    print("""Importing the htsmodels module. L. Roque. 
    Forecasting models for Hierarchical Time Series Forecasting Algorithms.\n""")


import numpy as np
from sklearn.metrics import mean_squared_error


def mase(n, seas, h, y, f):
    return np.mean(((n - seas) / h
                    * (np.sum(np.abs(y[n:n + h, :] - f), axis=0)
                       / np.sum(np.abs(y[seas:n, :] - y[:n - seas, :]), axis=0))))


def calculate_metrics(pred_samples,
                      groups):
    pred_s0 = pred_samples.shape[0]
    pred_s1 = pred_samples.shape[1]
    pred_s2 = pred_samples.shape[2]

    pred_samples = pred_samples.reshape(pred_s0, pred_s1 * pred_s2, order='F')

    seasonality = groups['seasonality']
    h = groups['h']

    n = groups['predict']['n']
    s = groups['predict']['s']
    y_f = groups['predict']['data'].reshape(s, n).T

    y_all_g = {}
    f_all_g = {}

    mase_ = {}
    rmse_ = {}

    # Bottom
    y_all_g['bottom'] = y_f
    f_all_g['bottom'] = np.mean(pred_samples, axis=0).reshape(s, n).T[n - h:n, :]

    mase_['bottom'] = np.round(mase(n=n - h,
                                    seas=seasonality,
                                    h=h,
                                    y=y_f,
                                    f=np.mean(pred_samples, axis=0).reshape(s, n).T[n - h:n, :]), 3)
    rmse_['bottom'] = np.round(
        mean_squared_error(y_f[n - h:n, :], np.mean(pred_samples, axis=0).reshape(s, n).T[n - h:n, :], squared=False),
        3)

    # Total
    y_all_g['total'] = np.sum(y_f, axis=1).reshape(-1, 1)
    f_all_g['total'] = np.sum(np.mean(pred_samples, axis=0).reshape(s, n).T[n - h:n, :], axis=1).reshape(-1, 1)

    mase_['total'] = np.round(mase(n=n - h,
                                   seas=seasonality,
                                   h=h,
                                   y=np.sum(y_f, axis=1).reshape(-1, 1),
                                   f=np.sum(np.mean(pred_samples, axis=0).reshape(s, n).T[n - h:n, :], axis=1).reshape(
                                       -1, 1))
                              , 3)
    rmse_['total'] = np.round(mean_squared_error(np.sum(y_f, axis=1).reshape(-1, 1)[n - h:n, :],
                                                 np.sum(np.mean(pred_samples, axis=0).reshape(s, n).T[n - h:n, :],
                                                        axis=1).reshape(-1, 1),
                                                 squared=False), 3)

    # Groups
    idx_dict_new = {}
    for group in list(groups['predict']['groups_names'].keys()):
        y_g = np.zeros((groups['predict']['n'], groups['predict']['groups_names'][group].shape[0]))
        f_g = np.zeros((h, groups['predict']['groups_names'][group].shape[0]))

        for idx, name in enumerate(groups['predict']['groups_names'][group]):
            g_n = groups['predict']['groups_n'][group]

            idx_dict_new[name] = np.where(groups['predict']['groups_idx'][group] == idx, 1, 0)

            y_g[:, idx] = np.sum(idx_dict_new[name] * y_f, axis=1)
            f_g[:, idx] = np.sum(idx_dict_new[name] * np.mean(pred_samples, axis=0).reshape(s, n).T, axis=1)[n - h:n]

        y_all_g[group] = np.sum(y_g, axis=1).reshape(-1, 1)
        f_all_g[group] = np.sum(f_g, axis=1).reshape(-1, 1)

        mase_[group] = np.round(mase(n=n - h,
                                     seas=seasonality,
                                     h=h,
                                     y=y_g,
                                     f=f_g)
                                , 3)

        rmse_[group] = np.round(mean_squared_error(y_g[n - h:n, :], f_g, squared=False), 3)

    # All
    y_all = np.concatenate([y_all_g[x] for x in y_all_g], 1)
    f_all = np.concatenate([f_all_g[x] for x in f_all_g], 1)

    mase_['all'] = np.round(mase(n=n - h,
                                 seas=seasonality,
                                 h=h,
                                 y=y_all,
                                 f=f_all), 3)
    rmse_['all'] = np.round(mean_squared_error(y_all[n - h:n, :], f_all, squared=False), 3)

    results = {}
    results['mase'] = mase_
    results['rmse'] = rmse_
    return results





import unittest
import tsaugmentation as tsag
from htsmodels.models.mint import MinT
import shutil


class TestModel(unittest.TestCase):

    def setUp(self):
        self.data = tsag.preprocessing.PreprocessDatasets('m5', top=100, test_size=1000).apply_preprocess()
        self.n = self.data['predict']['n']
        self.s = self.data['train']['s']
        shutil.rmtree("./data/original_datasets")
        self.mint = MinT(dataset='m5', groups=self.data)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree("./results")

    def test_correct_train(self):
        model = self.mint.train()
        self.assertIsNotNone(model)

    def test_predict_shape(self):
        forecasts = self.mint.train()
        res = self.mint.results(forecasts)
        self.assertTrue(res.shape == (1, self.n, self.s))



import pandas as pd
import rpy2.robjects as robjects
from rpy2.robjects import pandas2ri
import rpy2.robjects as ro
from rpy2.robjects.conversion import localconverter
import numpy as np
from htsmodels.results.calculate_metrics import calculate_metrics
import pickle
from pathlib import Path


class MinT:

    def __init__(self, dataset, groups, aggregate_key=None, input_dir='./'):
        self.dataset = dataset
        self.groups = groups
        # Ensure that keys are capitalized for every dataset
        self.groups['train']['groups_names'] = self._get_capitalized_keys(self.groups['train']['groups_names'])
        self.groups['train']['groups_idx'] = self._get_capitalized_keys(self.groups['train']['groups_idx'])
        self.groups['train']['groups_n'] = self._get_capitalized_keys(self.groups['train']['groups_n'])
        self.input_dir = input_dir
        self._create_directories()
        dict_groups = {k.capitalize(): np.tile(groups['train']['groups_names'][k][groups['train']['groups_idx'][k]],
                                               (groups['predict']['n'], 1)).T.reshape(-1, ) for k in
                       [k for k, v in groups['train']['groups_n'].items()]}
        groups['predict']['data_matrix'][:groups['train']['n'],:] = groups['train']['data']
        dict_groups['Count'] = groups['predict']['data_matrix'].T.reshape(-1,)
        dict_groups['Date'] = np.tile(np.array(groups['dates']), (groups['train']['s'],))
        self.df = pd.DataFrame(dict_groups)

        time_interval = (self.groups['dates'][1] - self.groups['dates'][0]).days
        if time_interval < 8:
            self.time_int = 'week'
        elif time_interval < 32:
            self.time_int = 'month'
        elif time_interval < 93:
            self.time_int = 'quarter'
        elif time_interval < 367:
            self.time_int = 'year'

        self.last_train_date = (self.groups['dates'][self.groups['train']['n']-1]).strftime("%Y-%m-%d")
        if aggregate_key:
            self.aggregate_key = aggregate_key
        else:
            # Default is to assume that there is a group structure (e.g. State * Gender)
            # and no direct hierarchy (e.g. State / Region)
            self.aggregate_key = ' * '.join([k.capitalize() for k in self.groups['train']['groups_names']])

    def _create_directories(self):
        # Create directory to store results if does not exist
        Path(f'{self.input_dir}results').mkdir(parents=True, exist_ok=True)

    def train(self):
        robjects.r('''
            library('fpp3')
            results_fn <- function(df, time, h, string_aggregate, start_predict_date) {
              if (time == 'quarter') {
                fn = yearquarter
              } else if (time == 'month') {
                fn = yearmonth
              } else if (time == 'week') {
                fn = yearweek
              } else if (time == 'year') {
                fn = year
              } 
              data <- df %>%
                mutate(Time = fn(Date)) %>%
                select(-Date) %>%
                as_tsibble(key = colnames(df)[2:length(colnames(df))-2], index = Time) %>%
                relocate(Time)
              
              data_gts <- data %>%
                aggregate_key(.spec = !!rlang::parse_expr(string_aggregate), Count = sum(Count))
              
              fit <- data_gts %>%
                filter(Time <= fn(as.Date(start_predict_date))) %>%
                model(base = ETS(Count)) %>%
                reconcile(
                  bottom_up = bottom_up(base),
                  MinT = min_trace(base, method = "mint_shrink")
                )
              fc <- fit %>% forecast(h = h)
              
              fc_csv = fc %>% 
                as_tibble %>% 
                filter(.model=='MinT') %>% 
                select(-Count) %>% 
                mutate(.mean=.mean) %>%
                mutate(.mean=(sprintf("%0.2f", .mean))) %>%
                rename(time=Time) %>%
                lapply(as.character) %>% 
                data.frame(stringsAsFactors=FALSE)
              
              return (fc_csv)
            }
        ''')
        function_r = robjects.globalenv['results_fn']
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_r = ro.conversion.py2rpy(self.df)

        df_result_r = function_r(df_r,
                                 self.time_int,
                                 self.groups['h'],
                                 self.aggregate_key,
                                 self.last_train_date)
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_result = ro.conversion.rpy2py(df_result_r)
        df_result[['.mean']] = df_result[['.mean']].astype('float')
        return df_result

    @staticmethod
    def sort_dataframe(df, columns_to_sort, express_filter_from_lines, indexes_dict):
        # filter from lines
        df = df[(df != express_filter_from_lines).all(axis=1)]
        # change the columns to categories data type
        for col in columns_to_sort:
            df = df.copy()
            df[col] = df[col].astype("category")
            df[col].cat.set_categories(indexes_dict[col], inplace=True)
        # sort values for all of the columns
        df = df.sort_values(columns_to_sort)
        df = df.reset_index().drop('index', axis=1)

        return df

    @staticmethod
    def _get_capitalized_keys(dict_to_capitalize):
        indexes_dict = {}
        for k, v in dict_to_capitalize.items():
            # ensure that groups names are capitalized
            indexes_dict[k.capitalize()] = v
        return indexes_dict

    def results(self, pred_mint):
        indexes_dict = self.groups['train']['groups_names']
        columns_to_sort = list(indexes_dict.keys())
        # Dataframe received from R needs to be in the same order as the original dataset
        df = self.sort_dataframe(pred_mint, columns_to_sort, '<aggregated>', indexes_dict)

        # Assert order is correct between original dataset and predictions
        # Should be in the unit tests folder, but to avoid messing with the API
        # it is here for now
        for group in indexes_dict.keys():
            np.testing.assert_array_equal(list(pd.unique(df[group])),
                                          list(self.groups['train']['groups_names'][group]))

        pred = df['.mean'].to_numpy().reshape(self.groups['train']['s'], self.groups['h']).T
        pred_complete = np.concatenate((np.zeros((self.groups['train']['n'],
                                                  self.groups['train']['s'])), pred), axis=0)[np.newaxis, :, :]
        return pred_complete

    def store_metrics(self, res):
        with open(f'{self.input_dir}results/results_gp_cov_{self.dataset}.pickle', 'wb') as handle:
            pickle.dump(res, handle, pickle.HIGHEST_PROTOCOL)

    def metrics(self, mean):
        res = calculate_metrics(mean, self.groups)
        return res



import unittest
import tsaugmentation as tsag
from htsmodels.models.deepar import DeepAR
import shutil


class TestModel(unittest.TestCase):

    def setUp(self):
        self.data = tsag.preprocessing.PreprocessDatasets('prison').apply_preprocess()
        self.n = self.data['predict']['n']
        self.s = self.data['train']['s']
        shutil.rmtree("./data/original_datasets")
        self.deepar = DeepAR(dataset='prison', groups=self.data)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree("./results")

    def test_correct_train(self):
        model = self.deepar.train(epochs=10)
        self.assertIsNotNone(model)

    def test_predict_shape(self):
        model = self.deepar.train(epochs=10)
        forecasts = self.deepar.predict(model)
        res = self.deepar.results(forecasts)
        self.assertTrue(res.shape == (100, self.n, self.s))

    def test_results_interval(self):
        model = self.deepar.train(epochs=10)
        forecasts = self.deepar.predict(model)
        results = self.deepar.results(forecasts)
        res = self.deepar.metrics(results)
        self.assertLess(res['mase']['bottom'], 2.8)



from setuptools import setup, find_packages
from os.path import dirname, join, realpath
import re

DESCRIPTION = 'Forecasting algorithms for hierarchical time series'
LONG_DESCRIPTION = 'A package that includes methods to forecast ' \
                   'hierarchical time series forecasting'

PROJECT_ROOT = dirname(realpath(__file__))
REQUIREMENTS_FILE = join(PROJECT_ROOT, "requirements.txt")
with open(REQUIREMENTS_FILE) as f:
    install_reqs = f.read().splitlines()


def get_version():
    VERSIONFILE = join("htsmodels", "__init__.py")
    lines = open(VERSIONFILE).readlines()
    version_regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
    for line in lines:
        mo = re.search(version_regex, line, re.M)
        if mo:
            return mo.group(1)
    raise RuntimeError(f"Unable to find version in {VERSIONFILE}.")

# Setting up
setup(
    name="htsmodels",
    version=get_version(),
    author="Luis Roque",
    author_email="<roque0luis@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=install_reqs,
    keywords=['python', 'time series', 'hierarchical', 'forecasting', 'htsmodels', 'machine learning'],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ]
)